#include "AIGroundNetworkLoader.h"

#include "AINetworkArc.h"
#include "AINetworkNode.h"
#include "StartupLocationMapObject.h"

AIGroundNetworkLoader::AIGroundNetworkLoader(Airport::SmartRef airport):
        XMLVisitor(),
        _airport(airport)
{
}

AIGroundNetworkLoader::~AIGroundNetworkLoader() {
}

void  AIGroundNetworkLoader::startXML() {
        _nodeIndexMap.clear();
        _arcBySourceNodeMap.clear();
        /* First delete everything having to do with the ground network */
        MapObjectContainer::SmartRef groundNetworkLayer=_airport->GetGroundNetworkLayer();
        MapObjectList objects;
        
        groundNetworkLayer->GetObjectsOfClass(objects, CLASSINFO(AINetworkVertex));
        
        MapObjectList::Node* objectNode;
        for (objectNode=objects.GetFirst();objectNode!=NULL;objectNode=objectNode->GetNext()) {
                MapObject::SmartRef obj=objectNode->GetData();
                obj->RemoveFromContainer();
        }
}

void  AIGroundNetworkLoader::endXML() {
        ArcSetMap::iterator it;
        
        for (it=_arcBySourceNodeMap.begin();it!=_arcBySourceNodeMap.end();++it) {
                unsigned int sourceIdx=it->first;
                ArcSet& arcSet=it->second;
                ArcSet::iterator nodeIt;
                
                // TODO: Check that this really exists
                AINetworkVertex::SmartRef source=_nodeIndexMap[sourceIdx];
                for (nodeIt=arcSet.begin();nodeIt!=arcSet.end();++nodeIt) {
                        unsigned int targetIdx=nodeIt->target;
                        bool isBidirectional=false;
                        Arc inverseArc(targetIdx,sourceIdx,nodeIt->isPushBackRoute,wxT(""));
                        if (_arcBySourceNodeMap[targetIdx].count(inverseArc)!=0) {
                                if (sourceIdx>targetIdx)
                                        continue; // insert bidirectionals only once
                                isBidirectional=true;
                        }
                        // TODO: Check that this really exists
                        AINetworkVertex::SmartRef target=_nodeIndexMap[targetIdx];
                        AINetworkArc::SmartRef arc=new AINetworkArc(_airport->GetTransactionManager(),
                                source,target,
                                isBidirectional,nodeIt->isPushBackRoute,nodeIt->name);
                        _airport->GetGroundNetworkLayer()->AttachObject(arc);
                }
        }
}

void  AIGroundNetworkLoader::startElement(const char * name, const XMLAttributes &atts) {
        if (name == string("Parking")) {
                parseParking(atts);
        } else if (name == string("node")) {
                parseNode(atts);
        } else if (name == string("arc")) {
                parseArc(atts);
        }
}

void AIGroundNetworkLoader::parseParking(const XMLAttributes& atts) {
        int index=0;
        Point3D pos;
        wxString name,number,airlinecodes;
        float heading=0.0,radius=17.3 /* Airbus A320 standard */;
        StartupLocationMapObject::GateType type=StartupLocationMapObject::GateType_Gate;
        
        string attname;
        for (int i=0;i<atts.size();i++) {
                attname=atts.getName(i);
                
                if (attname==string("index")) {
                        index=atoi(atts.getValue(i));
                } else if (attname==string("type")) {
                        string attval=atts.getValue(i);
                        if (attval==string("gate")) {
                                type=StartupLocationMapObject::GateType_Gate;
                        } else if (attval==string("cargo")) {
                                type=StartupLocationMapObject::GateType_Cargo;
                        } else if (attval==string("ga")) {
                                type=StartupLocationMapObject::GateType_GA;
                        } else if (attval==string("mil-fighter")) {
                                type=StartupLocationMapObject::GateType_Mil_Fighter;
                        } else if (attval==string("mil-cargo")) {
                                type=StartupLocationMapObject::GateType_Mil_Cargo;
                        }
                } else if (attname==string("name")) {
                        name=wxString(atts.getValue(i),wxConvUTF8);
                } else if (attname==string("number")) {
                        number=wxString(atts.getValue(i),wxConvUTF8);
                } else if (attname==string("lat")) {
                        pos.setlat(processPosition(atts.getValue(i)));
                } else if (attname==string("lon")) {
                        pos.setlon(processPosition(atts.getValue(i)));
                } else if (attname==string("heading")) {
                        heading=atof(atts.getValue(i));
                } else if (attname==string("radius")) {
                        radius=atof(atts.getValue(i));
                } else if (attname==string("airlineCodes")) {
                        airlinecodes=wxString(atts.getValue(i),wxConvUTF8);
                }
        }
        StartupLocationMapObject::SmartRef startupLocation=
                new StartupLocationMapObject(_airport,
                                             pos,
                                             heading,
                                             name,number,
                                             type,
                                             airlinecodes,
                                             radius);
        _airport->GetGroundNetworkLayer()->AttachObject(startupLocation);
        _nodeIndexMap[index]=startupLocation;
}

void AIGroundNetworkLoader::parseNode(const XMLAttributes& atts) {
        int index=0;
        Point3D pos;
        bool isOnRunway=false;
        AINetworkNode::HoldPointType holdPointType=AINetworkNode::HoldPointType_None;
        
        string attname;
        for (int i=0;i<atts.size();i++) {
                attname=atts.getName(i);
                
                if (attname==string("index")) {
                        index=atoi(atts.getValue(i));
                } else if (attname==string("lat")) {
                        pos.setlat(processPosition(atts.getValue(i)));
                } else if (attname==string("lon")) {
                        pos.setlon(processPosition(atts.getValue(i)));
                } else if (attname==string("holdPointType")) {
                        string attval=atts.getValue(i);
                        if (attval==string("none")) {
                                holdPointType=AINetworkNode::HoldPointType_None;
                        } else if (attval==string("normal")) {
                                holdPointType=AINetworkNode::HoldPointType_Normal;
                        } else if (attval==string("CAT II/III")) {
                                holdPointType=AINetworkNode::HoldPointType_CAT_II_III;
                        } else if (attval==string("PushBack")) {
                                holdPointType=AINetworkNode::HoldPointType_PushBack;
                        }
                } else if (attname==string("isOnRunway")) {
                        isOnRunway=(bool)atoi(atts.getValue(i));
                }
                
        }
        AINetworkNode::SmartRef node=new AINetworkNode(_airport->GetTransactionManager(),pos,isOnRunway,holdPointType);
        _airport->GetGroundNetworkLayer()->AttachObject(node);
        _nodeIndexMap[index]=node;
}

void AIGroundNetworkLoader::parseArc(const XMLAttributes& atts) {
        int begin,end;
        bool isPushBackRoute=false;
        wxString name=wxT("Route");
        
        string attname;
        for (int i=0;i<atts.size();i++) {
                attname=atts.getName(i);
                
                if (attname==string("begin")) {
                        begin=atoi(atts.getValue(i));
                } else if (attname==string("end")) {
                        end=atoi(atts.getValue(i));
                } else if (attname==string("isPushBackRoute")) {
                        isPushBackRoute=(bool)atoi(atts.getValue(i));
                } else if (attname==string("name")) {
                        name=wxString(atts.getValue(i),wxConvUTF8);
                }
        }
        if (begin==end) {
                std::cout << "self-referential arc on " << begin << ", ignoring\n";
                return;
        }
        _arcBySourceNodeMap[begin].insert(Arc(begin,end,isPushBackRoute,name));
}

double AIGroundNetworkLoader::processPosition(const string &pos)
{
  string prefix;
  string subs;
  string degree;
  string decimal;
  int sign = 1;
  double value;
  subs = pos;
  prefix= subs.substr(0,1);
  if (prefix == string("S") || (prefix == string("W")))
    sign = -1;
  subs    = subs.substr(1, subs.length());
  degree  = subs.substr(0, subs.find(" ",0));
  decimal = subs.substr(subs.find(" ",0), subs.length());
  
	      
  //cerr << sign << " "<< degree << " " << decimal << endl;
  value = sign * (atof(degree.c_str()) + atof(decimal.c_str())/60.0);
  //cerr << value <<endl;
  //exit(1);
  return value;
}

