#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "XPlaneWriter.h"

#ifdef _MSC_VER
#  include <Main/TaxiDraw.h>
#endif

#include <Airport/Airport.h>
#include <Airport/BeaconMapObject.h>
#include <Airport/PavementMapObject.h>
#include <Airport/RunwayMapObject.h>
#include <Airport/TaxiwayMapObject.h>
#include <Airport/TowerMapObject.h>
#include <Airport/WindsockMapObject.h>

#include <AI/StartupLocationMapObject.h>

#include <Map/MapObjectContainer.h>

#include <iomanip>

#if _MSC_VER >= 1400
#  define strtok_r strtok_s
#elif _MSC_VER >= 1300
#  define strtok_r(a,b,c) strtok(a,b)
#endif

IMPLEMENT_CLASS(XPlaneWriter,wxObject);

XPlaneWriter::XPlaneWriter(std::ostream& outfile):
        _outfile(outfile),
        _version(0)
{
}

XPlaneWriter::~XPlaneWriter() {
}

void XPlaneWriter::writeHeader() {
        // TODO: find out whether we write PC or MAC convention at EOL
        _outfile << "I\n";
        _outfile << _version << " Version. Created with " << PACKAGE << " " << VERSION << "\n";
        _outfile << "\n";
}

void XPlaneWriter::writeAirport(Airport::SmartRef airport) {
        // write airport header
        writeAirportHeader(airport);
        
        MapObjectList list;
        MapObjectList::Node* listNode;

        /* NOTE: The X-Plane-format expects the objects in top-to-bottom order,
         *       but the layers store them in bottom-to-top order, therefore
         *       we need to swap the order.
         */
        // write runways
        list.Clear();
        airport->GetRunwayLayer()->GetObjectsOfClass(list, CLASSINFO(RunwayMapObject));
        for (listNode=list.GetLast();listNode;listNode=listNode->GetPrevious()) {
                writeRunway(wxDynamicCast(listNode->GetData(),RunwayMapObject));
        }

        // write taxiways
        list.Clear();
        airport->GetTaxiwayLayer()->GetObjectsOfClass(list, CLASSINFO(TaxiwayMapObject));
        for (listNode=list.GetLast();listNode;listNode=listNode->GetPrevious()) {
                writeTaxiway(wxDynamicCast(listNode->GetData(),TaxiwayMapObject));
        }
        
        // write tower viewpoints
        list.Clear();
        airport->GetObjectsLayer()->GetObjectsOfClass(list, CLASSINFO(TowerMapObject));
        for (listNode=list.GetLast();listNode;listNode=listNode->GetPrevious()) {
                writeTowerViewpoint(wxDynamicCast(listNode->GetData(),TowerMapObject));
        }

        // write startup locations
        list.Clear();
        airport->GetObjectsLayer()->GetObjectsOfClass(list, CLASSINFO(StartupLocationMapObject));
        for (listNode=list.GetLast();listNode;listNode=listNode->GetPrevious()) {
                writeStartupLocation(wxDynamicCast(listNode->GetData(),StartupLocationMapObject));
        }

        // write beacons
        list.Clear();
        airport->GetObjectsLayer()->GetObjectsOfClass(list, CLASSINFO(BeaconMapObject));
        for (listNode=list.GetLast();listNode;listNode=listNode->GetPrevious()) {
                writeBeacon(wxDynamicCast(listNode->GetData(),BeaconMapObject));
        }

        // write windsocks
        list.Clear();
        airport->GetObjectsLayer()->GetObjectsOfClass(list, CLASSINFO(WindsockMapObject));
        for (listNode=list.GetLast();listNode;listNode=listNode->GetPrevious()) {
                writeWindsock(wxDynamicCast(listNode->GetData(),WindsockMapObject));
        }

        // write frequencies
        // CHECK: Do we have to write them in groups of equal types?
        const FrequencyEntryArray& frequencies=airport->GetFrequencies();
        for (size_t i=0;i<frequencies.GetCount();i++) {
                writeFrequency(frequencies.Item(i));
        }
        
        _outfile << "\n";
}

void XPlaneWriter::writeFooter() {
        _outfile << "99\n";
}

void XPlaneWriter::mergeAirport(std::istream& infile, Airport::SmartRef airport) {
        char buf[2000],*token,*saveptr;
        char linestr[2000];
        bool skipLines=false;
        bool found=false;
        
        const char space[]=" \n\t\r";
        
        while (!infile.getline(buf,sizeof(buf)).fail()) {
                strcpy(linestr,buf);
                token=strtok_r(buf,space,&saveptr);
                
                if (token==NULL) { // empty line
                        if (!skipLines) _outfile << linestr << "\n";
                        continue;
                }
                
                if (!strcmp(token,"99")) {
                        if (!found) writeAirport(airport);
                        _outfile << linestr << "\n";
                        break; // last line
                }
                
                if (!strcmp(token,"1") ||
                           !strcmp(token,"16") ||
                           !strcmp(token,"17")) {
                        /* Found an airport line, skip to ICAO ID */
                        skipLines=false;
                        if (!strtok_r(NULL,space,&saveptr) ||
                            !strtok_r(NULL,space,&saveptr) ||
                            !strtok_r(NULL,space,&saveptr) ||
                            !(token=strtok_r(NULL,space,&saveptr)))
                        {
                                _outfile << linestr << "\n";
                                continue;
                        }
                        if (airport->GetICAOID()==wxString(token,wxConvUTF8)) {
                                if (!found) writeAirport(airport);
                                found=true;
                                skipLines=true;
                        }
                }
                
                if (!skipLines) {
                        _outfile << linestr << "\n";
                }
                
                if (!strcmp(token,"715")) {
                        _version=715;
                } else if (!strcmp(token,"810")) {
                        _version=810;
                } else if (!strcmp(token,"840")) {
                        _version=840;
                }
        }
}

void XPlaneWriter::writeAirportHeader(Airport::SmartRef airport) {
        int type;
        
        switch (airport->GetType()) {
        case Airport::AirportType_Land:
                type=1;
                break;
        case Airport::AirportType_Sea:
                type=16;
                break;
        case Airport::AirportType_Heliport:
                type=17;
                break;
        }
        
        _outfile << std::setw(2) << std::left
                 << type << " "
                 << std::setw(5) << std::right << airport->GetElevation() << " "
                 << std::resetiosflags(std::ios_base::adjustfield)
                 << (airport->HasControlTower()?1:0) << " "
                 << (airport->HasDefaultBuildings()?1:0) << " "
                 << airport->GetICAOID().mb_str(wxConvUTF8) << " "
                 << airport->GetName().mb_str(wxConvUTF8) << "\n";
}

void XPlaneWriter::writeRunway(RunwayMapObject::SmartRef rwyObject) {
        Point3D center=rwyObject->GetCenter();
        RunwayMapObject::RunwayEndData designatedEnd,oppositeEnd;
        
        designatedEnd=rwyObject->GetDesignatedEnd();
        oppositeEnd=rwyObject->GetOppositeEnd();
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << 10 << " " 
                 
                 << std::right << std::fixed
                 
                 << std::setw(10) << std::setprecision(6) << center.lat() << " "
                 << std::setw(11) << std::setprecision(6) << center.lon() << " "
                 
                 << std::setw(3) << rwyObject->GetDesignation().mb_str(wxConvUTF8) << " "
                 
                 << std::setw(7) << std::setprecision(3) << rwyObject->GetHeading() << " "
                 
                 << std::setw(5) << (unsigned int)rwyObject->GetLength() << " "
                 
                 << std::setfill('0')
                 
                 << std::setw(4) << designatedEnd.displacedThreshold << "."
                 << std::setw(4) << oppositeEnd.displacedThreshold << " "
                 
                 << std::setw(4) << designatedEnd.stopwayLength << "."
                 << std::setw(4) << oppositeEnd.stopwayLength << " "
                 
                 << std::setfill(' ')
                 
                 << std::setw(4) << (unsigned int)rwyObject->GetWidth() << " "
                 
                 << GetCodeForVAPL(designatedEnd.visualApproachPathLighting)
                 << GetCodeForRWYL(designatedEnd.runwayLighting)
                 << GetCodeForAPPL(designatedEnd.approachLighting)
                 << GetCodeForVAPL(oppositeEnd.visualApproachPathLighting)
                 << GetCodeForRWYL(oppositeEnd.runwayLighting)
                 << GetCodeForAPPL(oppositeEnd.approachLighting)
                 << " "
                 
                 << std::setw(2) << std::setfill('0')
                 << GetCodeForSurface(rwyObject->GetSurface()) << " "
                 << std::setfill(' ')
                 
                 << GetCodeForShoulder(rwyObject->GetShoulder()) << " "
                 
                 << GetCodeForMarkings(rwyObject->GetMarkings()) << " "
                 
                 << std::setw(4) << std::setprecision(2) << rwyObject->GetSmoothness() << " "
                 
                 << std::setw(1) << (rwyObject->HasDistanceRemainingSigns()?1:0);
                 
        
        // VASI was added in version 810
        if (_version>=810) {
                _outfile << " "
                         << std::setfill('0')
                         
                         << std::setw(4) << (int)(designatedEnd.vasiAngle*100) << "."
                         << std::setw(4) << (int)(oppositeEnd.vasiAngle*100) << " "
                         
                         << std::setfill(' ');
        }
        
        _outfile << "\n";
}

void XPlaneWriter::writeTaxiway(TaxiwayMapObject::SmartRef twyObject) {
        Point3D center=twyObject->GetCenter();
        int lightCode;
        
        /*
         * Actually, we lose information here, but the X-Plane format
         * only uses code 6 and TerraGear interprets it as indicating both
         * centerline and edge lights.
         *
         * -- Ralf Gerlich, 2008-03-21
         */
        if (twyObject->HasCenterLights() || twyObject->HasEdgeLights())
                lightCode=6;
        else
                lightCode=1;
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << 10 << " " 
                 
                 << std::right << std::fixed
                 
                 << std::setw(10) << std::setprecision(6) << center.lat() << " "
                 << std::setw(11) << std::setprecision(6) << center.lon() << " "
                 
                 "xxx "
                 
                 << std::setw(7) << std::setprecision(3) << twyObject->GetHeading() << " "
                 
                 << std::setw(5) << (unsigned int)twyObject->GetLength() << " "
                 
                 << "0000.0000 0000.0000 "
                 
                 << std::setw(4) << (unsigned int)twyObject->GetWidth() << " "
                 
                 << "1" << lightCode << "11" << lightCode << "1 "
                 
                 << std::setw(2) << std::setfill('0')
                 << GetCodeForSurface(twyObject->GetSurface()) << " "
                 << std::setfill(' ')
                 
                 << "0 0 "
                 
                 << std::setw(4) << std::setprecision(2) << twyObject->GetSmoothness() << " "
                 
                 << "0 \n";
}

int XPlaneWriter::GetCodeForVAPL(RunwayMapObject::VisualApproachPathLighting vapl) {
        switch (vapl) {
        case RunwayMapObject::VAPL_None:
                return 1;
        case RunwayMapObject::VAPL_VASI:
                return 2;
        case RunwayMapObject::VAPL_PAPI:
                return 3;
        case RunwayMapObject::VAPL_SpaceShuttlePAPI:
                return 4;
        }
        return 1;
}

int XPlaneWriter::GetCodeForRWYL(RunwayMapObject::RunwayLighting rwyl) {
        switch (rwyl) {
        case RunwayMapObject::RWYL_None:
                return 1;
        case RunwayMapObject::RWYL_EdgeLights:
                return 2;
        case RunwayMapObject::RWYL_REIL:
                return 3;
        case RunwayMapObject::RWYL_CLL:
                return 4;
        case RunwayMapObject::RWYL_TDZ:
                return 5;
        }
        return 1;
}

int XPlaneWriter::GetCodeForAPPL(RunwayMapObject::ApproachLighting appl) {
        switch (appl) {
        case RunwayMapObject::APPL_None:
                return 1;
        case RunwayMapObject::APPL_SSALS:
                return 2;
        case RunwayMapObject::APPL_SALSF:
                return 3;
        case RunwayMapObject::APPL_ALSF_I:
                return 4;
        case RunwayMapObject::APPL_ALSF_II:
                return 5;
        case RunwayMapObject::APPL_ODALS:
                return 6;
        case RunwayMapObject::APPL_Calvert_1:
                return 7;
        case RunwayMapObject::APPL_Calvert_2_3:
                return 8;
        }
        return 1;
}

int XPlaneWriter::GetCodeForShoulder(RunwayMapObject::Shoulder shoulder) {
        switch (shoulder) {
        case RunwayMapObject::Shoulder_None:
                return 0;
        case RunwayMapObject::Shoulder_Asphalt:
                return 1;
        case RunwayMapObject::Shoulder_Concrete:
                return 2;
        }
        return 0;
}

int XPlaneWriter::GetCodeForMarkings(RunwayMapObject::Markings markings) {
        switch (markings) {
        case RunwayMapObject::Markings_None:
                return 0;
        case RunwayMapObject::Markings_Visual:
                return 1;
        case RunwayMapObject::Markings_NonPrecision:
                return 2;
        case RunwayMapObject::Markings_Precision:
                return 3;
        }
        return 0;
}

int XPlaneWriter::GetCodeForSurface(PavementMapObject::Surface surface) {
        switch (surface) {
        case PavementMapObject::Surface_Asphalt:
                return 1;
        case PavementMapObject::Surface_Concrete:
                return 2;
        case PavementMapObject::Surface_Turf:
                return 3;
        case PavementMapObject::Surface_Dirt:
                return 4;
        case PavementMapObject::Surface_Gravel:
                return 5;
        case PavementMapObject::Surface_Helipad_Asphalt:
                return 6;
        case PavementMapObject::Surface_Helipad_Concrete:
                return 7;
        case PavementMapObject::Surface_Helipad_Turf:
                return 8;
        case PavementMapObject::Surface_Helipad_Dirt:
                return 9;
        case PavementMapObject::Surface_Twy_Asphalt_yellow_hold:
                return 10;
        case PavementMapObject::Surface_Twy_Concrete_yellow_hold:
                return 11;
        case PavementMapObject::Surface_Lakebed:
                return 12;
        case PavementMapObject::Surface_Water:
                return 13;
        }
        return 1;
}

void XPlaneWriter::writeTowerViewpoint(TowerMapObject::SmartRef obj) {
        Point3D center=obj->GetPosition();
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << 14 << " " 
                 
                 << std::right << std::fixed
                 
                 << std::setw(10) << std::setprecision(6) << center.lat() << " "
                 << std::setw(11) << std::setprecision(6) << center.lon() << " "
                 
                 << std::setw(3) << std::setprecision(0) << obj->GetHeight();
        
        // "Draw tower" flag was added in version 715
        if (_version>=715) {
                _outfile << " "
                         
                         << std::setw(1) << (obj->GetDrawTower()?1:0);
        }
        
        _outfile << " "
                 
                 << obj->GetName().mb_str(wxConvUTF8) << "\n";
}

void XPlaneWriter::writeStartupLocation(StartupLocationMapObject::SmartRef obj) {
        Point3D center=obj->GetPosition();
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << 15 << " " 
                 
                 << std::right << std::fixed
                 
                 << std::setw(10) << std::setprecision(6) << center.lat() << " "
                 << std::setw(11) << std::setprecision(6) << center.lon() << " "
                 
                 << std::setw(7) << std::setprecision(3) << obj->GetHeading() << " "
                 
                 << obj->GetName().mb_str(wxConvUTF8) << "\n";
}

void XPlaneWriter::writeBeacon(BeaconMapObject::SmartRef obj) {
        Point3D center=obj->GetPosition();
        int beaconCode;
        
        switch (obj->GetBeaconType()) {
        case BeaconMapObject::Beacon_None:
                beaconCode=0;
                break;
        case BeaconMapObject::Beacon_WhiteGreen_Flash:
                beaconCode=1;
                break;
        case BeaconMapObject::Beacon_WhiteYellow_Flash:
                beaconCode=2;
                break;
        case BeaconMapObject::Beacon_GreenYellowWhite_Flash:
                beaconCode=3;
                break;
        case BeaconMapObject::Beacon_WhiteWhiteGreen_Flash:
                beaconCode=4;
                break;
        case BeaconMapObject::Beacon_White_Strobe:
                beaconCode=5;
                break;
        }
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << 18 << " " 
                 
                 << std::right << std::fixed
                 
                 << std::setw(10) << std::setprecision(6) << center.lat() << " "
                 << std::setw(11) << std::setprecision(6) << center.lon() << " "
                 
                 << std::setw(1) << beaconCode << " "
                 
                 << obj->GetName().mb_str(wxConvUTF8) << "\n";
}

void XPlaneWriter::writeWindsock(WindsockMapObject::SmartRef obj) {
        Point3D center=obj->GetPosition();
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << 19 << " " 
                 
                 << std::right << std::fixed
                 
                 << std::setw(10) << std::setprecision(6) << center.lat() << " "
                 << std::setw(11) << std::setprecision(6) << center.lon() << " "
                 
                 << std::setw(1) << (obj->IsLighted()?1:0) << " "
                 
                 << obj->GetName().mb_str(wxConvUTF8) << "\n";
}

void XPlaneWriter::writeFrequency(FrequencyEntry freq) {
        int freqCode;
        
        switch (freq.GetType()) {
        case FrequencyEntry::FrequencyType_AWOS:
                freqCode=50;
                break;
        case FrequencyEntry::FrequencyType_UNICOM:
                freqCode=51;
                break;
        case FrequencyEntry::FrequencyType_Clearance:
                freqCode=52;
                break;
        case FrequencyEntry::FrequencyType_Ground:
                freqCode=53;
                break;
        case FrequencyEntry::FrequencyType_Tower:
                freqCode=54;
                break;
        case FrequencyEntry::FrequencyType_Approach:
                freqCode=55;
                break;
        case FrequencyEntry::FrequencyType_Departure:
                freqCode=56;
                break;
        }
        
        _outfile << std::setw(0) << std::setprecision(100)
                 << freqCode<< " " 
                 
                 << freq.GetFrequency().mb_str(wxConvUTF8) << " "
                 
                 << freq.GetName().mb_str(wxConvUTF8) << "\n";
}

