#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "USGSFetcher.h"
#include <wx/config.h>
#include <wx/wfstream.h>

#include <Projection/UTMProjection.h>

#include <wx/listimpl.cpp>
WX_DEFINE_LIST(USGSProgressListenerList);

IMPLEMENT_CLASS(USGSFetcher,wxEvtHandler);
IMPLEMENT_CLASS(USGSTileLoadedEvent,wxEvent);

DEFINE_EVENT_TYPE(wxEVT_USGS_TILE_LOADED);

USGSFetcher* USGSFetcher::_singletonInstance=NULL;

BEGIN_EVENT_TABLE(USGSFetcher,wxEvtHandler)
        EVT_USGS_TILE_LOADED(USGSFetcher::OnTileLoaded)
END_EVENT_TABLE()

static size_t USGSWrite(void *ptr, size_t size, size_t nmemb, void *stream) {
        wxOutputStream* realStream=(wxOutputStream*)stream;
        realStream->Write(ptr,size*nmemb);
        return realStream->LastWrite();
}

USGSFetcher* USGSFetcher::GetInstance() {
        if (!_singletonInstance) {
                _singletonInstance=new USGSFetcher();
                _singletonInstance->Create();
                _singletonInstance->Run();
        }
        return _singletonInstance;
}

USGSFetcher::USGSFetcher():
        _pendingRequests(0),
        _totalRequests(0)
{
        _queue.DeleteContents(false);
}

USGSFetcher::~USGSFetcher() {
}

void* USGSFetcher::Entry() {
        while (true) {
                _queueSema.Wait();
                _queueAccess.Enter();
                MapUSGSImageList::Node* reqNode;
                reqNode=_queue.GetFirst();
                MapUSGSImage *image=reqNode->GetData();
                _queue.DeleteNode(reqNode);
                
                ProcessRequest(image);
                _queueAccess.Leave();
                Yield();
        }
        return NULL;
}

void USGSFetcher::AddRequests(const MapUSGSImageList& images) {
        _queueAccess.Enter();
        MapUSGSImageList::Node* node;
        for (node=images.GetFirst();node;node=node->GetNext()) {
                MapUSGSImage* image=node->GetData();
                image->TransactionObject::Ref();
                if (::wxFileExists(image->GetFilename())) {
                        _queue.Insert(image);
                } else {
                        _queue.Append(image);
                }
                _queueSema.Post();
                _totalRequests++;
                _pendingRequests++;
        }
        FireUpdateProgress(_pendingRequests,_totalRequests);
        _queueAccess.Leave();
}

void USGSFetcher::CancelRequests(const MapUSGSImageList& images) {
        _queueAccess.Enter();
        MapUSGSImageList::Node* node;
        for (node=images.GetFirst();node;node=node->GetNext()) {
                MapUSGSImage* image=node->GetData();
                MapUSGSImageList::Node* queuenode=_queue.Find(image);
                if (!queuenode)
                        continue;
                _queue.DeleteNode(queuenode);
                image->TransactionObject::UnRef();
                _queueSema.Wait();
                _totalRequests--;
                _pendingRequests--;
        }
        FireUpdateProgress(_pendingRequests,_totalRequests);
        _queueAccess.Leave();
}

void USGSFetcher::CancelPendingRequests() {
        _queueAccess.Enter();
        MapUSGSImageList::Node* node;
        for (node=_queue.GetFirst();node;node=node->GetNext()) {
                MapUSGSImage* image=node->GetData();
                image->TransactionObject::UnRef();
                _queueSema.Wait();
                _totalRequests--;
                _pendingRequests--;
        }
        _queue.Clear();
        FireUpdateProgress(_pendingRequests,_totalRequests);
        _queueAccess.Leave();
}

void USGSFetcher::RegisterProgressListener(USGSProgressListener* listener) {
        _progressListeners.Append(listener);
}

void USGSFetcher::UnregisterProgressListener(USGSProgressListener* listener) {
        _progressListeners.DeleteObject(listener);
}

void USGSFetcher::FireUpdateProgress(unsigned int pending, unsigned int total) {
        USGSProgressListenerList::Node *node;
        for (node=_progressListeners.GetFirst();node;node=node->GetNext()) {
                node->GetData()->UpdateProgress(pending, total);
        }
}

void USGSFetcher::ProcessRequest(MapUSGSImage::SmartRef image) {
        const wxString& fullName=image->GetFilename();
        _pendingRequests--;
        if (!_pendingRequests)
                _totalRequests=0;
        if (!::wxFileExists(fullName)) {
                wxString partName=fullName+wxT("_part");
                wxFileOutputStream* stream=new wxFileOutputStream(partName);
                wxString request;
                request.Printf(wxT("http://terraserver-usa.com/tile.ashx?T=%d&S=%d&X=%d&Y=%d&Z=%d"),
                        image->GetType(),image->GetScale(),image->GetX(),image->GetY(),image->GetZone());
                wxCharBuffer urlStr=request.mb_str(wxConvUTF8);
                
                CURL* curl;
                curl=curl_easy_init();
                curl_easy_setopt(curl,CURLOPT_WRITEFUNCTION,USGSWrite);
                
                bool useProxy=false;
                
                wxConfigBase::Get()->Read(wxT("Connection/UseHttpProxy"), &useProxy);
                
                if (useProxy) {
                        wxString httpProxy,proxyPortStr;
                        long proxyPort;
                        wxConfigBase::Get()->Read(wxT("Connection/HttpProxyName"), &httpProxy);
                        wxConfigBase::Get()->Read(wxT("Connection/HttpProxyPort"), &proxyPortStr);
                        proxyPortStr.ToLong(&proxyPort);
                        curl_easy_setopt(curl, CURLOPT_PROXY, (const char*)httpProxy.mb_str(wxConvUTF8));
                        curl_easy_setopt(curl, CURLOPT_PROXYPORT, proxyPort);
                }
                
                curl_easy_setopt(curl,CURLOPT_URL,(const char*)urlStr);
                curl_easy_setopt(curl,CURLOPT_WRITEDATA,(void*)(wxOutputStream*)stream);
                CURLcode ret=curl_easy_perform(curl);
                if (ret!=0) {
                        std::cout << "USGSFetcher: Failed to fetch " << (const char*)urlStr << ":" << curl_easy_strerror(ret) << "\n";
                        ::wxRemoveFile(partName);
                        delete stream;
                        return;
                }
                delete stream;
                ::wxRenameFile(partName,fullName);
                
                curl_easy_cleanup(curl);
        } else {
                Sleep(50); // make sure the main thread can catch up
        }
        
        USGSTileLoadedEvent loadEvent(wxEVT_USGS_TILE_LOADED,image,_pendingRequests,_totalRequests);
        
        AddPendingEvent(loadEvent);
        
        image->TransactionObject::UnRef();
}

void USGSFetcher::OnTileLoaded(USGSTileLoadedEvent& evt) {
        MapUSGSImage* image=evt.GetImage();
        image->LoadImage();
        
        FireUpdateProgress(evt.GetPendingRequests(),evt.GetTotalRequests());
}

