#include "ProjectWizardOpenProjectPage.h"

#include "MapFrame.h"
#include "MapViewManager.h"

#include <IO/OldProjectLoader.h>
#include <Math/TDRectangle.h>
#include <Transactions/TransactionManager.h>

#include <wx/config.h>
#include <wx/filedlg.h>
#include <wx/filename.h>
#include <wx/progdlg.h>

#include <fstream>

IMPLEMENT_CLASS(ProjectWizardOpenProjectPage,wxWizardPageSimple);

enum {
        ProjectWizardOpenProjectPage_ID_Browse=wxID_HIGHEST,
        ProjectWizardOpenProjectPage_ID_RecentFiles
};

BEGIN_EVENT_TABLE(ProjectWizardOpenProjectPage,wxWizardPageSimple)
        EVT_BUTTON(ProjectWizardOpenProjectPage_ID_Browse,ProjectWizardOpenProjectPage::OnBrowse)
        EVT_LISTBOX(ProjectWizardOpenProjectPage_ID_RecentFiles,ProjectWizardOpenProjectPage::OnSelectRecent)
        EVT_WIZARD_PAGE_CHANGING(wxID_ANY,ProjectWizardOpenProjectPage::OnPageChanging)
END_EVENT_TABLE()

ProjectWizardOpenProjectPage::ProjectWizardOpenProjectPage(wxWizard* parent,
                     wxWizardPage* prev,
                     wxWizardPage* next,
                     const wxBitmap& bitmap,
                     const wxChar *resource):
             wxWizardPageSimple(parent,prev,next,bitmap,resource)
{
        wxBoxSizer* topSizer=new wxBoxSizer(wxVERTICAL);
        
        wxStaticText* text;
        
        text=new wxStaticText(this,wxID_ANY,wxT("Enter the name of the file"));
        topSizer->Add(text,wxALIGN_LEFT);
        
        wxBoxSizer* browseSizer=new wxBoxSizer(wxHORIZONTAL);
        
        _fileName=new wxTextCtrl(this,wxID_ANY,wxT(""));
        wxSize ctrlSize=_fileName->GetSize();
        ctrlSize.SetWidth(300);
        _fileName->SetMinSize(ctrlSize);
        browseSizer->Add(_fileName,wxSizerFlags(1).Expand());
        
        browseSizer->Add(new wxButton(this,
                                      ProjectWizardOpenProjectPage_ID_Browse,
                                      wxT("...")),
                        wxSizerFlags(0).Right());
        
        topSizer->Add(browseSizer,wxSizerFlags(0).Expand());
        
        _recentFiles=new wxListBox(this,
                                   ProjectWizardOpenProjectPage_ID_RecentFiles);
        
        topSizer->Add(_recentFiles,wxSizerFlags(1).Expand());
        
        SetSizer(topSizer);
        
        topSizer->SetSizeHints(this);
        
        Fit();
}

ProjectWizardOpenProjectPage::~ProjectWizardOpenProjectPage() {
}

bool ProjectWizardOpenProjectPage::TransferDataToWindow() {
        if (!wxWizardPageSimple::TransferDataToWindow())
                return false;
        // populate the list of recently opened files
        wxArrayString recentlyOpened;
        _recentFiles->Clear();
        for (int i=0;;i++) {
                wxString regPath;
                regPath.Printf(wxT("History/Path/recent%d"),i);
                if (!wxConfigBase::Get()->HasEntry(regPath))
                        break;
                wxString fullPath=wxConfigBase::Get()->Read(regPath);
                wxFileName filename=wxFileName(fullPath);
                recentlyOpened.Add(filename.GetName());
        }
        _recentFiles->Append(recentlyOpened);
        
        return true;
}

void ProjectWizardOpenProjectPage::OnPageChanging(wxWizardEvent& evt) {
        if (!evt.GetDirection())
                return;
        wxString project_file=_fileName->GetValue();
        if (project_file.IsEmpty() || !::wxFileExists(project_file)) {
                ::wxMessageBox(wxT("The given project file does not exist.")
                               wxT("Choose an existing filename."),
                               wxT("File does not exist"),
                               wxOK|wxICON_ERROR);
                evt.Veto();
                return;
        }
	

        wxFileName project_filename=wxFileName(project_file);
        TransactionManager* tmgr=new TransactionManager();
        tmgr->Hold();
        
        TDRectangle bounds;
        
        std::ifstream infile(project_file.mb_str(*wxConvFileName));
        OldProjectLoader loader(tmgr,infile);
        
        Airport::SmartRef airport;
        MapObjectContainer::SmartRef backgroundImageLayer;
        
        backgroundImageLayer=new MapObjectContainer(tmgr,wxT("Background images"));
        wxProgressDialog progressDialog(wxT("Loading Project..."),
                                        wxT("Loading project file                 "),
                                        0, // no progress gauge
                                        this);
        wxString projectionName;
        airport=loader.LoadProject(&progressDialog,projectionName,backgroundImageLayer);
        if (!airport) {
                ::wxMessageBox(wxT("The given project file could not be loaded. ")
                               wxT("Perhaps the project file is corrupted.\n")
                               wxT("Filename: ")+project_file,
                               wxT("Unable to load project file"),
                               wxOK|wxICON_ERROR);
                tmgr->Unhold();
                evt.Veto();
                return;
        }
        
        MapProjection *proj=NULL;
        if (!projectionName.IsEmpty()) {
		proj=MapProjectionDescriptor::CreateInstanceForName(projectionName);
		if (!proj) {
			::wxMessageBox(wxT("Unknown projection:")+projectionName+wxT("\n")
				       wxT("Will use standard projection"),
				       wxT("Warning"),
				       wxOK|wxICON_HAND);
		}
	}
        
        MapFrame* newFrame=new MapFrame(airport,backgroundImageLayer);
        
        tmgr->Unhold();
        newFrame->SetProjectFilename(project_file);
        MapEditor* newEditor=newFrame->GetEditor();
        
        bounds=newEditor->GetGeographicBounds();
        Point3D center=bounds.GetCenter();
        
        /*
         * We need to initialise the standard projection anyway
         * to make ZoomToGeo() work.
         */
        newEditor->GetProjection()->Init(center);
        
        if (!proj) {
                proj=newEditor->GetProjection();
        } else {
                proj->Init(center);
        }
        
        newEditor->ZoomToGeo(bounds);
        
        newEditor->SetProjection(proj);
        
        newEditor->GetTransactionManager()->Clear();
        
        progressDialog.Show(false);
        
        newFrame->Show();
        MapViewManager::GetInstance()->ActivateView(newFrame);
        
        wxConfigBase::Get()->Write(wxT("History/Paths/last-open-path"),project_filename.GetPath());
        
        // update the list of recently opened files
        wxArrayString recentlyOpened;
        for (int i=0;;i++) {
                wxString regPath;
                regPath.Printf(wxT("History/Path/recent%d"),i);
                if (!wxConfigBase::Get()->HasEntry(regPath))
                        break;
                wxString fullPath=wxConfigBase::Get()->Read(regPath);
                wxConfigBase::Get()->DeleteEntry(regPath);
                if (fullPath.Cmp(project_file)==0)
                        continue;
                if (!::wxFileExists(fullPath))
                        continue;
                if (recentlyOpened.GetCount()<9)
                        recentlyOpened.Add(fullPath);
        }
        recentlyOpened.Insert(project_file,0);
        for (size_t i=0;i<recentlyOpened.GetCount();i++) {
                wxString regPath;
                regPath.Printf(wxT("History/Path/recent%d"),i);
                wxString fullPath=recentlyOpened.Item(i);
                wxConfigBase::Get()->Write(regPath,fullPath);
        }
}

void ProjectWizardOpenProjectPage::OnBrowse(wxCommandEvent&) {
        wxString fullPath=_fileName->GetValue();
        wxString path,name;
        
        if (path.IsEmpty()) {
                path=wxConfigBase::Get()->Read(wxT("History/Paths/last-open-path"),wxFileName::GetCwd());
        } else {
                wxFileName filename=wxFileName(fullPath);
                path=filename.GetPath();
                name=filename.GetName();
        }
        
        path=::wxFileSelector(wxT("Select a project file"),
                              path,
                              name,
                              wxT("tpj"),
                              wxT("TaxiDraw project files (*.tpj)|*.tpj"),
                              0,this);
        if (path.empty())
                return;
        _fileName->SetValue(path);
}

void ProjectWizardOpenProjectPage::OnSelectRecent(wxCommandEvent&) {
        wxArrayInt selections;
#       ifdef __WXDEBUG__
                // agh, otherwise the compiler complains about unused var
                int cnt=
#       endif
        _recentFiles->GetSelections(selections);
        wxASSERT(cnt==1);
        int i=selections.Item(0);
        wxString regPath;
        regPath.Printf(wxT("History/Path/recent%d"),i);
        wxString fullPath=wxConfigBase::Get()->Read(regPath);
        _fileName->SetValue(fullPath);
}
