#ifndef MapCoordinateSystem_h
#define MapCoordinateSystem_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include <Math/point3d.h>
#include <Math/TDRectangle.h>
#include <algorithm>

/**
 * \brief A MapCoordinateSystem encapsulates the transformation between device
 * (canvas) coordinate system and the logical coordinate system.
 *
 * The coordinate system is defined by an origin in device coordinates and
 * an origin in logical coordinates as well as a logical scale, which is
 * expressed as "device units per logical units".
 *
 * The mapping is performed so that the logical origin maps onto the device
 * origin and 1 logical unit corresponds to logicalScale device units.
 *
 * Note that MapCoordinateSystem does not perform projection. Chain application
 * of a MapProjection instance and a MapCoordinateSystem instance to perform
 * conversion from geographic coordinates to device coordinates.
 *
 * See \ref coordinate_systems for a more detailed description of the coordinate
 * systems.
 */
class MapCoordinateSystem {
        friend class MapEditor;
public:
        MapCoordinateSystem():
                _deviceOriginX(0),
                _deviceOriginY(0),
                _logicalOriginX(0),
                _logicalOriginY(0),
                _logicalScale(1.0) {}
        
        /**
         * \brief Convert a logical X coordinate to a device X coordinate.
         */
        wxCoord LogicalToDeviceX(double lx) {
                double offx=(lx-_logicalOriginX)*_logicalScale;
                offx+=(offx>0?0.5:-0.5);
                return _deviceOriginX+(wxCoord)offx;
        }
        
        /**
         * \brief Convert a logical Y coordinate to a device Y coordinate.
         */
        wxCoord LogicalToDeviceY(double ly) {
                double offy=(ly-_logicalOriginY)*_logicalScale;
                offy+=(offy>0?0.5:-0.5);
                return _deviceOriginY-(wxCoord)offy;
        }
        
        /**
         * \brief Convert a logical X offset to a device X offset.
         *
         * This method differs from LogicalToDeviceX in that (0,0) in
         * the logical coordinate system maps to (0,0) in the device coordinate
         * system.
         *
         * It can be used to convert relative offsets.
         */
        wxCoord LogicalToDeviceRelX(double lx) {
                double offx=lx*_logicalScale;
                offx+=(offx>0?0.5:-0.5);
                return (wxCoord)offx;
        }
        
        
        /**
         * \brief Convert a logical Y offset to a device Y offset.
         *
         * This method differs from LogicalToDeviceY in that (0,0) in
         * the logical coordinate system maps to (0,0) in the device coordinate
         * system.
         *
         * It can be used to convert relative offsets.
         */
        wxCoord LogicalToDeviceRelY(double ly) {
                double offy=ly*_logicalScale;
                offy+=(offy>0?0.5:-0.5);
                return -(wxCoord)offy;
        }
        
        /**
         * \brief Convert a logical X size to a device X size.
         *
         * This method converts the given logical size to a device size
         * and clips it to the range [devMin, devMax].
         *
         * This is used, e.g., when determining the size of the cirles for
         * point features in TaxiDraw maps.
         *
         * \param devMin        Minimum size in device coordinate system
         * \param devMax        Maximum size in device coordinate syste
         * \param logStd        Size in logical coordinate system
         */
        wxCoord BoundedDeviceSizeX(wxCoord devMin, wxCoord devMax,
                                   double logStd) {
                /* sizes are always >=0, in contrast to offsets */
                wxCoord devStd=(wxCoord)(fabs(logStd)*_logicalScale+0.5);
                return std::min(std::max(devStd,devMin),devMax);
        }
        
        /**
         * \brief Convert a logical Y size to a device Y size.
         *
         * This method converts the given logical size to a device size
         * and clips it to the range [devMin, devMax].
         *
         * This is used, e.g., when determining the size of the cirles for
         * point features in TaxiDraw maps.
         *
         * \param devMin        Minimum size in device coordinate system
         * \param devMax        Maximum size in device coordinate syste
         * \param logStd        Size in logical coordinate system
         */
        wxCoord BoundedDeviceSizeY(wxCoord devMin, wxCoord devMax,
                                   double logStd) {
                /* sizes are always >=0, in contrast to offsets */
                wxCoord devStd=(wxCoord)(fabs(logStd)*_logicalScale+0.5);
                return std::min(std::max(devStd,devMin),devMax);
        }
        
        /**
         * \brief Convenience method to convert a logical point to a device
         * point.
         */
        wxPoint LogicalToDevice(Point3D p) {
                return wxPoint(LogicalToDeviceX(p.x()),
                               LogicalToDeviceY(p.y()));
        }
        
        /**
         * \brief Convenience method to convert a logical rectangle to a device
         * rectangle.
         */
        wxRect LogicalToDevice(TDRectangle r) {
                wxPoint p1=LogicalToDevice(r.GetTopLeft());
                wxPoint p2=LogicalToDevice(r.GetBottomRight());
                return wxRect(p1,p2);
        }
        
        /**
         * \brief Convenience method to convert a logical offset to a device
         * offset.
         */
        wxSize LogicalToDeviceRel(Point3D deltap) {
                return wxSize(LogicalToDeviceRelX(deltap.x()),
                              LogicalToDeviceRelY(deltap.y()));
        }
        
        /**
         * \brief Convert a device X coordinate to a logical X coordinate.
         */
        double DeviceToLogicalX(wxCoord dx) {
                double offx=(dx-_deviceOriginX)/_logicalScale;
                return offx+_logicalOriginX;
        }
        
        /**
         * \brief Convert a device Y coordinate to a logical Y coordinate.
         */
        double DeviceToLogicalY(wxCoord dy) {
                double offy=(_deviceOriginY-dy)/_logicalScale;
                return offy+_logicalOriginY;
        }
        
        /**
         * \brief Convert a device X offset to a logical X offset.
         *
         * This method differs from DeviceToLogicalX in that (0,0) in
         * the device coordinate system maps to (0,0) in the logical coordinate
         * system.
         *
         * It can be used to convert relative offsets.
         */
        double DeviceToLogicalRelX(wxCoord dx) {
                double offx=dx/_logicalScale;
                return offx;
        }
        
        /**
         * \brief Convert a device Y offset to a logical Y offset.
         *
         * This method differs from DeviceToLogicalY in that (0,0) in
         * the device coordinate system maps to (0,0) in the logical coordinate
         * system.
         *
         * It can be used to convert relative offsets.
         */
        double DeviceToLogicalRelY(wxCoord dy) {
                double offy=-dy/_logicalScale;
                return offy;
        }
        
        /**
         * \brief Find a logical X size that will be displayed as a device size
         * within given bounds.
         *
         * This method finds a logical size that will map to a device size
         * in the range [devMin, devMax]. It uses the given logical size
         * as a hint.
         *
         * This is used, e.g., when determining the width of the drag bars
         * on the border of pavement objects in TaxiDraw maps.
         *
         * \param devMin        Minimum size in device coordinate system
         * \param devMax        Maximum size in device coordinate syste
         * \param logStd        Size in logical coordinate system
         */
        double BoundedLogicalSizeX(wxCoord devMin, wxCoord devMax,
                                   double logStd) {
                return std::min(std::max(logStd,devMin/_logicalScale),
                                devMax/_logicalScale);
	}
        
        /**
         * \brief Find a logical Y size that will be displayed as a device size
         * within given bounds.
         *
         * This method finds a logical size that will map to a device size
         * in the range [devMin, devMax]. It uses the given logical size
         * as a hint.
         *
         * This is used, e.g., when determining the width of the drag bars
         * on the border of pavement objects in TaxiDraw maps.
         *
         * \param devMin        Minimum size in device coordinate system
         * \param devMax        Maximum size in device coordinate syste
         * \param logStd        Size in logical coordinate system
         */
        double BoundedLogicalSizeY(wxCoord devMin, wxCoord devMax,
                                   double logStd) {
                return std::min(std::max(logStd,devMin/_logicalScale),
                                devMax/_logicalScale);
        }

        /**
         * \brief Convenience method to convert a device point to a logical
         * point.
         */
        Point3D DeviceToLogical(wxPoint p) {
                return Point3D(DeviceToLogicalX(p.x),
                               DeviceToLogicalY(p.y),
                               0);
        }
        
        /**
         * \brief Convenience method to convert a device rectangle to a logical
         * rectangle.
         */
        TDRectangle DeviceToLogical(wxRect r) {
                Point3D p1=DeviceToLogical(wxPoint(r.GetTop(),r.GetLeft()));
                Point3D p2=DeviceToLogical(wxPoint(r.GetBottom(),r.GetRight()));
                return TDRectangle(p1,p2);
        }
        
        /**
         * \brief Convenience method to convert a device offset to a logical
         * offset.
         */
        Point3D DeviceToLogicalRel(wxSize deltap) {
                return Point3D(DeviceToLogicalRelX(deltap.x),
                               DeviceToLogicalRelY(deltap.y),
                               0);
        }
        
        /**
         * \brief Retrieve the device origin.
         *
         * The device origin is mapped to the logical origin on conversion.
         */
        void GetDeviceOrigin(wxCoord& x, wxCoord& y) {
                x=_deviceOriginX;
                y=_deviceOriginY;
        }
        
        /**
         * \brief Retrieve the logical origin.
         *
         * The logical origin is mapped to the device origin on conversion.
         */
        void GetLogicalOrigin(double& x, double& y) {
                x=_logicalOriginX;
                y=_logicalOriginY;
        }
        
        /**
         * \brief Retrieve the logical scale.
         *
         * The logical scale is the number of device units a logical unit
         * is mapped to.
         */
        double GetLogicalScale() {
                return _logicalScale;
        }
protected:
        /* keep these protected so that only friend classes can modify
         * the coordinate system.
         * Currently the only friend class is MapEditor, meaning that
         * the editor is informed of any changes as a matter of fact.
         * As soon as the list of friends is to be extended, we should
         * subclass this so the coordinate system can only be modified
         * by its owner and not by any friend class.
         */
        void SetDeviceOrigin(wxCoord x, wxCoord y) {
                _deviceOriginX=x;
                _deviceOriginY=y;
        }
        
        void SetLogicalOrigin(double x, double y) {
                _logicalOriginX=x;
                _logicalOriginY=y;
        }
        
        void SetLogicalScale(double scale) {
                _logicalScale=scale;
        }
        
        wxCoord _deviceOriginX,_deviceOriginY;
        double _logicalOriginX,_logicalOriginY;
        double _logicalScale;
};

#endif
