#ifndef MapSelection_h
#define MapSelection_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include <wx/propgrid/propgrid.h>
#include <Math/TDRectangle.h>
#include "MapCoordinateSystem.h"
#include "MapDraggable.h"
#include "MapObject.h"
#include "MapObjectContainer.h"

#include <Transactions/TransactionObjectChangeListener.h>


class MapSelectionManager;
class MapSelectionHandle;

/**
 * \brief A MapSelection encapsulates the fact that a specific MapObject is selected.
 *
 * A MapSelection also provides handles to the user to manipulate the MapObject
 * via the selection.
 *
 * See \ref selections for more information on how to implement a selection.
 */
class MapSelection: public wxObject,
                public TransactionObjectChangeListener,
                public MapDraggable {
        
        DECLARE_ABSTRACT_CLASS(MapSelection)
        
        friend class MapSelectionManager;
        
public:
        /**
         * \brief Construct a selection in the given selection manager with
         * the given content.
         *
         * The selection automatically attaches itself to the
         * selection manager and registers the selection as
         * TransactionObjectLifetimeObserver on the content object.
         *
         * May only be called from appropriate content object.
         */
	MapSelection(MapSelectionManager* manager, MapObject* content);

        /**
         * \brief Destruct the selection.
         *
         * May only be called from UnRef.
         */
        virtual ~MapSelection();
        
        /**
         * \brief Activate a selection.
         *
         * This and DeactivateSelection() are called by the MapSelectionManager
         * when a selection is activated resp. deactivated.
         *
         * The current activation status of the selection is determined and if
         * it has changed, either DoActivateSelection() or DoDeactivateSelection()
         * are called.
         *
         * \note
         * Rationale: As selection objects are not necessarily destroyed on
         * deselection, they need to explicitly unregister from all services.
         * Otherwise they would react to these services even when not active
         * (e.g., because they are contained in a previous transaction).
         * 
         * \par
         * Currently there is no more elegant way of implementing that.
         *
         * \sa DeactivateSelection()
         */
        void ActivateSelection();
        
        /**
         * \brief Deactivate a selection.
         *
         * \sa ActivateSelection()
         */
        void DeactivateSelection();
        
        /**
         * \brief Determine whether the selection intersects with a given point.
         *
         * Checks whether the selected object or one of the selection handles
         * contains the given point.
         *
         * \param p     The point in logical coordinates.
         *
         * \return true if and only if the point hits the selection.
         */
        virtual bool Hit(Point3D p);
        
        /**
         * \brief Determine whether a given point hits one of the selection
         *        handles.
         *
         * \param p     The point in logical coordinates.
         *
         * \return the hit MapSelectionHandle, if any is hit, NULL otherwise.
         */
        virtual MapSelectionHandle* HitHandle(Point3D p);
        
        /**
         * \brief Get the bounds of this selection including its handles
         * in device coordinates.
         *
         * This should return a bounding box that exactly contains the area
         * overwritten when painting the selection in OnPaint().
         */
        virtual wxRect GetDeviceBounds() const;
        
        /**
         * \brief Get the bounds of this selection including its handles
         * in logical coordinates.
         */
        virtual TDRectangle GetLocalBounds() const;

        /**
         * \brief Provide menu actions for this selection type.
         *
         * Add actions to the actions menu, but only if this class is
         * a superclass of the given target class or equals the target class.
         *
         * Make sure to call the superclass method in any case.
         *
         * Event handlers for the given command events need to be defined.
         *
         * \param actionMenu    The menu to which to add actions.
         * \param targetClass   The target class for which to provide actions.
         */
        virtual void ProvideActionMenuForClass(wxMenu& actionMenu, wxClassInfo* targetClass);

        virtual void StartDrag(Point3D) {
        }
        
        virtual void Drag(Point3D) {
        }
        
        virtual wxString EndDrag() {
                return wxT("Dragging object");
        }
        
        /**
         * \brief Schedule this selection for repaint.
         */
        void Damage() const;
        
        /**
         * \brief Paint this selection.
         */
        virtual void Paint(wxDC& dc);
        
        /**
         * \brief Activate selection handles.
         */
        void ActivateHandles() {
                if (_handlesActive)
                        return;
                Damage(); // redraw old bounds
                _handlesActive=true;
                Damage(); // redraw new bounds
        }
        
        /**
         * \brief Check whether handles are active.
         */
        bool HandlesActive() const {
                return _handlesActive;
        }
        
        /**
         * \brief Deactivate selection handles.
         */
        void DeactivateHandles() {
                if (!_handlesActive)
                        return;
                Damage(); // redraw old bounds
                _handlesActive=false;
                Damage(); // redraw new bounds
        }
        
        /**
         * \brief Populate a property grid with properties for this class.
         *
         * Add properties for this class to the given grid, but only if this
         * class is a superclass of the given target class.
         *
         * Make sure to call the superclass method in any case.
         *
         * This way only those properties are added which are supported
         * by all currently present selections.
         *
         * \param propgrid      The property grid to populate
         * \param targetClass   The class for which to generate properties.
         */
        virtual void ProvidePropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass);
        
        /**
         * \brief Update the values in the properties grid from the content object.
         *
         * Update the properties for this class from the content object, but
         * only if this class is a superclass of the given target class.
         *
         * Make sure to call the superclass method in any case.
         *
         * \sa ProvidePropertiesForClass() MarkDifferingPropertiesForClass()
         */
        virtual void UpdatePropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass);
        
        /** \brief Mark properties differing among the selected objects.
         *
         * Each call is passed the first selection object, which is to be used
         * to determine whether the value for a property differs from the value
         * of the content of the first selection. If so, the respective property
         * is to be marked as unspecified.
         *
         * Only process your own properties if this class is a superclass
         * of the given target class.
         *
         * Make sure to call the superclass method in any case.
         *
         * \note Marking properties as unspecified currently inhibits change
         *       events for such properties, so that these properties cannot
         *       be edited by the user. A solution is being investigated.
         *
         * \sa ProvidePropertiesForClass() UpdatePropertiesForClass()
         */
        virtual void MarkDifferingPropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass, MapSelection* firstSelection);
        
        /**
         * \brief Write back a property value to the content object.
         *
         * This method is called when a property in the property grid was
         * changed by the user.
         */
        virtual void UpdateObjectFromProperty(wxPGProperty* property);
        
        virtual void BeforeObjectChange(TransactionObject* object);
        
        virtual void AfterObjectChange(TransactionObject* object);
        
        /**
         * \brief Get the MapSelectionManager for this selection.
         */
        MapSelectionManager* GetSelectionManager() const { return _selectionManager; }
        
        /**
         * \brief Convenience method to get the MapEditor for this selection.
         */
        MapEditor* GetEditor() const;
        
        /**
         * \brief Return the standard size for selection handle boxes in device coordinates.
         */
        wxSize GetHandleBoxDeviceSize() const;
        
        /**
         * \brief Return the standard size for selection handle boxes in logical coordinates.
         */
        Point3D GetHandleBoxLogicalSize() const;
        
        /**
         * \brief Get the content object of this selection.
         */
        MapObject* GetContent() const { return _content; }
        
        void Ref() { _refCount++; }
        void UnRef() { if (!--_refCount) delete this; }
        
        static wxColour Selection_Handle_Colour;
        static wxColour Selection_Marking_Colour;
protected:
        /**
         * \brief Actually activate the selection.
         *
         * This is called by ActivateSelection() when the
         * activation status is changed to 'active'.
         *
         * Perform here all necessary actions to connect the selections
         * to the rest of the system, e.g., by registering listeners.
         *
         * When overriding this method, the method on the super class should be
         * called first.
         *
         * \sa ActivateSelection() DeactivateSelection() DoDeactivateSelection()
         */
        virtual void DoActivateSelection();

        /**
         * \brief Actually deactivate the selection.
         *
         * This is called by DectivateSelection() when the
         * activation status is changed to 'inactive'.
         *
         * Undo the operations performed in DoActivateSelection() here.
         *
         * When overriding this method, the method on the super class should be
         * called last.
         *
         * \sa ActivateSelection() DeactivateSelection() DoActivateSelection()
         */
        virtual void DoDeactivateSelection();

        MapSelectionManager* _selectionManager;
        // We need to hold a reference to the container if we want to access
        // the container of the content
        MapObjectContainer::SmartRef _contentContainer;
        MapObject::SmartRef _content;
        unsigned int _refCount;
        int _activationCount;
        bool _handlesActive;
};

WX_DECLARE_LIST(MapSelection,MapSelectionList);

#endif

